@@ -96,10 +96,13 @@ static void            terminal_app_action_reset            (GtkAction       *ac
                                                              TerminalApp     *app);
 static void            terminal_app_action_reset_and_clear  (GtkAction       *action,
                                                              TerminalApp     *app);
+static void            terminal_app_action_ctrl             (GtkAction       *action,
+                                                             TerminalApp     *app);
 static void            terminal_app_action_settings         (GtkAction       *action,
                                                              TerminalApp     *app);
 static void            terminal_app_action_quit             (GtkAction       *action,
                                                              TerminalApp     *app);
+
 struct _TerminalApp
 {
   HildonApp            __parent__;
@@ -129,6 +132,7 @@ static GtkActionEntry action_entries[] =
   { "terminal-menu", NULL, N_ ("_Terminal"), },
   { "reset", NULL, N_ ("_Reset"), NULL, NULL, G_CALLBACK (terminal_app_action_reset), },
   { "reset-and-clear", NULL, N_ ("Reset and C_lear"), NULL, NULL, G_CALLBACK (terminal_app_action_reset_and_clear), },
+  { "ctrl", NULL, N_ ("Send Ctrl-<some key>"), NULL, NULL, G_CALLBACK (terminal_app_action_ctrl), },
   { "settings", GTK_STOCK_PREFERENCES, N_ ("_Settings..."), NULL, NULL, G_CALLBACK (terminal_app_action_settings), },
   { "quit", GTK_STOCK_QUIT, N_ ("_Quit"), NULL, NULL, G_CALLBACK (terminal_app_action_quit), },
 };
@@ -165,6 +169,8 @@ static const gchar ui_description[] =
  "    <menuitem action='scrollbar'/>"
  "    <menuitem action='fullscreen'/>"
  "    <separator/>"
+ "    <menuitem action='ctrl'/>"
+ "    <separator/>"
  "    <menu action='font-menu'>"
  "      <menuitem action='10pt'/>"
  "      <menuitem action='12pt'/>"
@@ -182,6 +188,12 @@ static const gchar ui_description[] =
 G_DEFINE_TYPE (TerminalApp, terminal_app, HILDON_TYPE_APP);
 
 
+typedef struct {
+    GtkWidget *dialog;
+    gchar *ret;
+} ctrl_dialog_data;
+
+
 static void
 terminal_app_class_init (TerminalAppClass *klass)
 {
@@ -264,6 +276,7 @@ populate_menubar (TerminalApp *app, GtkAccelGroup *accelgroup)
   parent = attach_menu(menubar, actiongroup, accelgroup, "terminal-menu");
   attach_item(parent, actiongroup, accelgroup, "reset");
   attach_item(parent, actiongroup, accelgroup, "reset-and-clear");
+  attach_item(parent, actiongroup, accelgroup, "ctrl");
 
   /*attach_item(menubar, actiongroup, accelgroup, "settings");*/
 
@@ -372,6 +385,7 @@ terminal_app_init (TerminalApp *app)
   gint                 font_size;
   gboolean             scrollbar, reverse;
   GConfClient         *gconf_client;
+  GConfValue          *gconf_value;
 
 
   hildon_app_set_title(HILDON_APP(app), _("X Terminal"));
@@ -391,33 +405,47 @@ terminal_app_init (TerminalApp *app)
   if (error != NULL) {
       g_printerr("Unable to get font size from gconf: %s\n",
                  error->message);
-      font_size = OSSO_XTERM_DEFAULT_FONT_SIZE;
       g_error_free(error);
       error = NULL;
   }
 
-  scrollbar = gconf_client_get_bool(gconf_client,
-                                    OSSO_XTERM_GCONF_SCROLLBAR,
-                                    &error);
+  if (font_size == 0) {
+      font_size = OSSO_XTERM_DEFAULT_FONT_SIZE;
+  }
+
+  gconf_value = gconf_client_get(gconf_client,
+                                 OSSO_XTERM_GCONF_SCROLLBAR,
+                                 &error);
+
   if (error != NULL) {
       g_printerr("Unable to get scrollbar setting from gconf: %s\n",
                  error->message);
-      scrollbar = OSSO_XTERM_DEFAULT_SCROLLBAR;
       g_error_free(error);
       error = NULL;
   }
-  g_object_unref(G_OBJECT(gconf_client));
+  scrollbar = OSSO_XTERM_DEFAULT_SCROLLBAR;
+  if (gconf_value) {
+          if (gconf_value->type == GCONF_VALUE_BOOL)
+                  scrollbar = gconf_value_get_bool(gconf_value);
+          gconf_value_free(gconf_value);
+  }
 
-  reverse = gconf_client_get_bool(gconf_client,
-                                  OSSO_XTERM_GCONF_REVERSE,
-                                  &error);
+  gconf_value = gconf_client_get(gconf_client,
+                                 OSSO_XTERM_GCONF_REVERSE,
+                                 &error);
   if (error != NULL) {
       g_printerr("Unable to get reverse setting from gconf: %s\n",
                  error->message);
-      reverse = OSSO_XTERM_DEFAULT_REVERSE;
       g_error_free(error);
       error = NULL;
   }
+  reverse = OSSO_XTERM_DEFAULT_REVERSE;
+  if (gconf_value) {
+          if (gconf_value->type == GCONF_VALUE_BOOL)
+                  reverse = gconf_value_get_bool(gconf_value);
+          gconf_value_free(gconf_value);
+  }
+
   g_object_unref(G_OBJECT(gconf_client));
 
 
@@ -841,6 +869,100 @@ terminal_app_action_reset_and_clear (GtkAction    *action,
   terminal_widget_reset (active, TRUE);
 }
 
+static void
+terminal_app_send_ctrl_key(TerminalApp *app,
+                           const char *str)
+{
+  GdkEventKey *key;
+
+  key = (GdkEventKey *) gdk_event_new(GDK_KEY_PRESS);
+
+  key->window = GTK_WIDGET(app)->window;
+  key->time = GDK_CURRENT_TIME;
+  key->state = GDK_CONTROL_MASK;
+  key->keyval = gdk_keyval_from_name(str);
+  gdk_event_put ((GdkEvent *) key);
+
+  key->type = GDK_KEY_RELEASE;
+  key->state |= GDK_RELEASE_MASK;
+  gdk_event_put ((GdkEvent *) key);
+
+  gdk_event_free((GdkEvent *) key);
+}
+
+static gboolean ctrl_dialog_focus(GtkWidget *dialog,
+                                  GdkEventFocus *event,
+                                  GtkIMContext *imctx)
+{
+  if (event->in) {
+    gtk_im_context_focus_in(imctx);
+    gtk_im_context_show(imctx);
+  } else
+    gtk_im_context_focus_out(imctx);
+  return FALSE;
+}
+
+static gboolean
+im_context_commit (GtkIMContext *ctx,
+                   const gchar *str,
+                   ctrl_dialog_data *data)
+{
+    if (strlen(str) > 0) {
+      data->ret = g_strdup(str);
+      gtk_dialog_response(GTK_DIALOG(data->dialog), GTK_RESPONSE_ACCEPT);
+    }
+
+    return TRUE;
+}
+
+static void
+terminal_app_action_ctrl (GtkAction    *action,
+                          TerminalApp  *app)
+{
+  ctrl_dialog_data *data;
+  GtkWidget *dialog, *label;
+  GtkIMContext *imctx;
+
+  dialog = gtk_dialog_new_with_buttons("Control",
+                                       GTK_WINDOW(app),
+                                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                       NULL);
+
+  imctx = gtk_im_multicontext_new();
+
+  data = g_new0(ctrl_dialog_data, 1);
+  data->dialog = dialog;
+  g_signal_connect(imctx, "commit", G_CALLBACK(im_context_commit), data);
+
+  label = gtk_label_new("Press a key (or cancel)");
+  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->vbox), label);
+
+  gtk_widget_show_all(dialog);
+
+  gtk_im_context_set_client_window(imctx, GTK_WIDGET(dialog)->window);
+
+  g_signal_connect( G_OBJECT(dialog), "focus-in-event",
+          G_CALLBACK(ctrl_dialog_focus), imctx);
+  g_signal_connect( G_OBJECT(dialog), "focus-out-event",
+          G_CALLBACK(ctrl_dialog_focus), imctx);
+
+  gtk_dialog_run(GTK_DIALOG(dialog));
+
+  gtk_widget_hide(dialog);
+  gtk_widget_destroy(dialog);
+
+  gtk_im_context_focus_out(imctx);
+  g_object_unref(G_OBJECT(imctx));
+
+  if (data->ret != NULL) {
+    terminal_app_send_ctrl_key(app, data->ret);
+    g_free(data->ret);
+  }
+
+  g_free(data);
+}
+
 
 static void
 terminal_app_action_settings (GtkAction    *action,
@@ -902,7 +1024,7 @@ terminal_app_add (TerminalApp    *app,
   g_return_if_fail (TERMINAL_IS_WIDGET (widget));
 
   popup = gtk_ui_manager_get_widget (app->ui_manager, "/popup-menu");
-  gtk_widget_tap_and_hold_setup(GTK_WIDGET(widget), popup, NULL,
+  gtk_widget_tap_and_hold_setup(GTK_WIDGET(widget->terminal), popup, NULL,
                                 GTK_TAP_AND_HOLD_NONE);
 
   active = terminal_app_get_active (app);
@@ -977,6 +1099,9 @@ terminal_app_launch (TerminalApp     *app,
   terminal_app_add (app, TERMINAL_WIDGET (terminal));
   terminal_widget_launch_child (TERMINAL_WIDGET (terminal));
 
+  /* Keep IM open on startup */
+  hildon_gtk_im_context_show(TERMINAL_WIDGET(terminal)->im_context);
+
   gtk_widget_show(GTK_WIDGET(app));
 
   return TRUE;
