@@ -61,25 +61,6 @@ enum
 };
 
 
-struct _TerminalWidget
-{
-  GtkHBox              __parent__;
-  GtkWidget           *terminal;
-  GtkWidget           *scrollbar;
-
-  GPid                 pid;
-  gchar               *working_directory;
-
-  gchar              **custom_command;
-  gchar               *custom_title;
-
-  GConfClient         *gconf_client;
-  guint                scrollbar_conid;
-  guint                font_size_conid;
-  guint                reverse_conid;
-};
-
-
 static void     terminal_widget_finalize                      (GObject          *object);
 static void     terminal_widget_get_property                  (GObject          *object,
                                                                guint             prop_id,
@@ -263,6 +244,28 @@ terminal_widget_class_init (TerminalWidgetClass *klass)
                   G_TYPE_NONE, 0);
 }
 
+static void
+terminal_widget_vte_focus_in_event (VteTerminal    *terminal,
+                                    GdkEventFocus *event,
+                                    TerminalWidget *widget)
+{
+  g_return_if_fail (VTE_IS_TERMINAL (terminal));
+  g_return_if_fail (TERMINAL_IS_WIDGET (widget));
+
+  gtk_im_context_focus_in (widget->im_context);
+}
+
+
+static void
+terminal_widget_vte_focus_out_event (VteTerminal    *terminal,
+                                    GdkEventFocus *event,
+                                    TerminalWidget *widget)
+{
+  g_return_if_fail (VTE_IS_TERMINAL (terminal));
+  g_return_if_fail (TERMINAL_IS_WIDGET (widget));
+
+  gtk_im_context_focus_out (widget->im_context);
+}
 
 static void
 terminal_widget_init (TerminalWidget *widget)
@@ -270,6 +273,7 @@ terminal_widget_init (TerminalWidget *widget)
   GError *err = NULL;
   gint font_size;
   gboolean scrollbar, reverse;
+  GConfValue *gconf_value;
 
   widget->working_directory = g_get_current_dir ();
   widget->custom_title = g_strdup ("");
@@ -315,7 +319,13 @@ terminal_widget_init (TerminalWidget *widget)
       g_clear_error(&err);
   }
 
+  widget->im_context = gtk_im_multicontext_new ();
   widget->terminal = vte_terminal_new ();
+
+  g_signal_connect (G_OBJECT (widget->terminal), "focus-in-event",
+                    G_CALLBACK (terminal_widget_vte_focus_in_event), widget);
+  g_signal_connect (G_OBJECT (widget->terminal), "focus-out-event",
+                    G_CALLBACK (terminal_widget_vte_focus_out_event), widget);
   g_signal_connect (G_OBJECT (widget->terminal), "child-exited",
                     G_CALLBACK (terminal_widget_vte_child_exited), widget);
   g_signal_connect (G_OBJECT (widget->terminal), "encoding-changed",
@@ -357,29 +367,42 @@ terminal_widget_init (TerminalWidget *widget)
   if (err != NULL) {
       g_printerr("Unable to get font size from gconf: %s\n",
                  err->message);
-      font_size = OSSO_XTERM_DEFAULT_FONT_SIZE;
       g_clear_error(&err);
   }
 
-  scrollbar = gconf_client_get_bool(widget->gconf_client,
-                                    OSSO_XTERM_GCONF_SCROLLBAR,
-                                    &err);
+  if (font_size == 0) {
+      font_size = OSSO_XTERM_DEFAULT_FONT_SIZE;
+  }
+
+  gconf_value = gconf_client_get(widget->gconf_client,
+                                 OSSO_XTERM_GCONF_SCROLLBAR,
+                                 &err);
   if (err != NULL) {
       g_printerr("Unable to get scrollbar setting from gconf: %s\n",
                  err->message);
-      scrollbar = OSSO_XTERM_DEFAULT_SCROLLBAR;
       g_clear_error(&err);
   }
+  scrollbar = OSSO_XTERM_DEFAULT_SCROLLBAR;
+  if (gconf_value) {
+          if (gconf_value->type == GCONF_VALUE_BOOL)
+                  scrollbar = gconf_value_get_bool(gconf_value);
+          gconf_value_free(gconf_value);
+  }
 
-  reverse = gconf_client_get_bool(widget->gconf_client,
-                                  OSSO_XTERM_GCONF_REVERSE,
-                                  &err);
+  gconf_value = gconf_client_get(widget->gconf_client,
+                                 OSSO_XTERM_GCONF_REVERSE,
+                                 &err);
   if (err != NULL) {
       g_printerr("Unable to get reverse setting from gconf: %s\n",
                  err->message);
-      reverse = OSSO_XTERM_DEFAULT_REVERSE;
       g_clear_error(&err);
   }
+  reverse = OSSO_XTERM_DEFAULT_REVERSE;
+  if (gconf_value) {
+          if (gconf_value->type == GCONF_VALUE_BOOL)
+                  reverse = gconf_value_get_bool(gconf_value);
+          gconf_value_free(gconf_value);
+  }
 
   terminal_widget_update_scrolling_bar(TERMINAL_WIDGET(widget), scrollbar);
   terminal_widget_update_font_size (TERMINAL_WIDGET(widget), font_size);
@@ -412,6 +435,8 @@ terminal_widget_finalize (GObject *object)
 
   g_object_unref(G_OBJECT(widget->gconf_client));
 
+  g_object_unref(widget->im_context);
+
   parent_class->finalize (object);
 }
 
@@ -640,7 +665,6 @@ terminal_widget_update_word_chars (TerminalWidget *widget)
 {
 }
 
-
 static void
 terminal_widget_vte_child_exited (VteTerminal    *terminal,
                                   TerminalWidget *widget)
@@ -781,11 +805,20 @@ terminal_widget_vte_button_press_event (VteTerminal    *terminal,
                                         GdkEventButton *event,
                                         TerminalWidget *widget)
 {
+  if (hildon_gtk_im_context_filter_event (widget->im_context, (GdkEvent*)event))
+    {
+      return TRUE;
+    }
+
   if (event->button == 3)
     {
       g_signal_emit (G_OBJECT (widget), widget_signals[CONTEXT_MENU], 0, event);
       return TRUE;
     }
+  else
+    {
+      hildon_gtk_im_context_show(widget->im_context);
+    }
 
   return FALSE;
 }
@@ -812,6 +845,8 @@ terminal_widget_vte_realize (VteTerminal    *terminal,
   vte_terminal_set_allow_bold (terminal, TRUE);
   terminal_widget_timer_background (TERMINAL_WIDGET (widget));
   terminal_widget_update_colors (TERMINAL_WIDGET (widget));
+
+  gtk_im_context_set_client_window (widget->im_context, widget->terminal->window);
 }
 
 
@@ -1295,7 +1330,6 @@ terminal_widget_reset (TerminalWidget *widget,
   vte_terminal_reset (VTE_TERMINAL (widget->terminal), TRUE, clear);
 }
 
-
 /**
  * terminal_widget_im_append_menuitems:
  * @widget    : A #TerminalWidget.
